#include <cmath>
#include "udx_api.hpp"
#include <sstream>
#include <iostream>
#include <map>
 
using namespace udx2;
 
struct DiscEntropy : Aggregate
{
  long m_cnt = 0; // Total events
  std::map<std::string, long> m_map;
 
  void accumulate(const Row& r)
  {
	// NOTE: Supply input as string even for numbers using SPARQL expression
	if(!r.getString(0).empty())
	{
  	m_map[std::string(r.getString(0).begin(), r.getString(0).end())]++;
  	m_cnt += 1;
	}
  }
 
  void save(Row& r)
  {
	if(m_cnt != 0)
	{
  	std::ostringstream out;
  	for(const auto& [key, val] : m_map)
     	out << key.size() << '~' << key << val << '~';
  	r.setLong(0, m_cnt)
   	.setString(1, out.str());
	}
  }
 
  void merge(const Row& r)
  {
	if(r.defined(0))
	{
  	std::istringstream in(r.getString(1).data());
  	std::string key;
  	long val;
  	int len;
  	char delimit;
  	while(in.good())
  	{
    	key.clear();
    	val = len = 0;
    	in >> len >> delimit;
    	if(in && len)
    	{
      	std::vector<char> tmp(len);
      	in.read(tmp.data(), len);
      	key.assign(tmp.data(), len);
      	in >> val >> delimit;
      	m_map[key] += val;
    	}
  	}
  	m_cnt += r.getLong(0);
	}
  }
 
  void result(Row& r)
  {
	if(m_cnt == 0)
  	azg_throw("http://example/aggregates#discentropy",": insufficient data");
	// Formula: Entropy = -Sum[p(x) * log(p(x))] for all classes of x events.
	double entropy = 0;
	double prob = 0;
	for(const auto& elem : m_map)
	{
  	prob = double(elem.second) / m_cnt;
  	entropy += prob * std::log2(prob);
	}
	r.setDouble(0, (entropy >= 0.0) ? entropy : -1 * entropy);
  }
};
 
extern "C" ExtensionFactory* udx_csi_statistics_discentropy() { return new FactoryFor<DiscEntropy>(); }

